/*-
 * Copyright (c) 1992, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 * Copyright (c) 1992, 1993, 1994, 1995, 1996
 *	Keith Bostic.  All rights reserved.
 *
 * See the LICENSE file for redistribution information.
 */

#include "config.h"

#ifndef lint
static const char sccsid[] = "@(#)v_xchar.c	10.9 (Berkeley) 10/23/96";
#endif /* not lint */

#include <sys/types.h>
#include <sys/queue.h>
#include <sys/time.h>

#include <bitstring.h>
#include <limits.h>
#include <stdio.h>

#include "../common/common.h"
#include "vi.h"

/*
 * v_xchar -- [buffer] [count]x
 *	Deletes the character(s) on which the cursor sits.
 *
 * PUBLIC: int v_xchar __P((SCR *, VICMD *));
 */
int
v_xchar(sp, vp)
	SCR *sp;
	VICMD *vp;
{
	size_t len;
	int isempty;
#ifdef MULTIBYTE
	CHAR_T *p;
	size_t lastlen;
#endif

#ifndef MULTIBYTE
	if (db_eget(sp, vp->m_start.lno, NULL, &len, &isempty))
#else
	if (db_eget(sp, vp->m_start.lno, &p, &len, &isempty))
#endif
	{
		if (isempty)
			goto nodel;
		return (1);
	}
	if (len == 0) {
nodel:		msgq(sp, M_BERR, "206|No characters to delete");
		return (1);
	}

	/*
	 * Delete from the cursor toward the end of line, w/o moving the
	 * cursor.
	 *
	 * !!!
	 * Note, "2x" at EOL isn't the same as "xx" because the left movement
	 * of the cursor as part of the 'x' command isn't taken into account.
	 * Historically correct.
	 */
	if (F_ISSET(vp, VC_C1SET)) {
#ifndef MULTIBYTE
		vp->m_stop.cno += vp->count - 1;
#else
		int i;
		for (i = 0; i < vp->count - 1; i++)
			vp->m_stop.cno += KEY_BLEN(sp, p[vp->m_stop.cno]);
#endif /*MULTIBYTE*/
	}
#ifndef MULTIBYTE
	if (vp->m_stop.cno >= len - 1) {
		vp->m_stop.cno = len - 1;
		vp->m_final.cno = vp->m_start.cno ? vp->m_start.cno - 1 : 0;
	}
#else
	lastlen = len - multi_back(sp, p, &p[len]);
	if (vp->m_stop.cno >= lastlen) {
		vp->m_stop.cno = lastlen;
		if (vp->m_start.cno) {
			vp->m_final.cno = vp->m_start.cno
				- multi_back(sp, p, &p[vp->m_start.cno]);
		} else
			vp->m_final.cno = 0;
	}
#endif /*MULTIBYTE*/
	else
		vp->m_final.cno = vp->m_start.cno;

	if (cut(sp,
	    F_ISSET(vp, VC_BUFFER) ? &vp->buffer : NULL,
	    &vp->m_start, &vp->m_stop, 0))
		return (1);
	return (del(sp, &vp->m_start, &vp->m_stop, 0));
}

/*
 * v_Xchar -- [buffer] [count]X
 *	Deletes the character(s) immediately before the current cursor
 *	position.
 *
 * PUBLIC: int v_Xchar __P((SCR *, VICMD *));
 */
int
v_Xchar(sp, vp)
	SCR *sp;
	VICMD *vp;
{
	u_long cnt;
#ifdef MULTIBYTE
	CHAR_T *p;
#endif

	if (vp->m_start.cno == 0) {
		v_sol(sp);
		return (1);
	}

#ifdef MULTIBYTE
	/* we need content of the line. */
	if (db_eget(sp, vp->m_start.lno, &p, NULL, NULL))
		return (1);
#endif

	cnt = F_ISSET(vp, VC_C1SET) ? vp->count : 1;
	if (cnt >= vp->m_start.cno)
		vp->m_start.cno = 0;
	else {
#ifndef MULTIBYTE
		vp->m_start.cno -= cnt;
#else
		int i;
		for (i = 0; i < cnt; i++)
			vp->m_start.cno -= multi_back(sp, p, &p[vp->m_start.cno]);
#endif /*MULTIBYTE*/
	}
#ifndef MULTIBYTE
	--vp->m_stop.cno;
#else
	vp->m_stop.cno -= multi_back(sp, p, &p[vp->m_stop.cno]);
#endif
	vp->m_final.cno = vp->m_start.cno;

	if (cut(sp,
	    F_ISSET(vp, VC_BUFFER) ? &vp->buffer : NULL,
	    &vp->m_start, &vp->m_stop, 0))
		return (1);
	return (del(sp, &vp->m_start, &vp->m_stop, 0));
}
